import json
import os
import shutil

import typing
import yaml


def make_dir(path):
    if not os.path.exists(path):
        print(" [*] Make directories : {}".format(path))
        os.makedirs(path)


def remove_dir(path):
    if os.path.exists(path):
        print(" [*] Removed: {}".format(path))
        shutil.rmtree(path)


def write_json(path, data):
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f, indent=4, sort_keys=True, ensure_ascii=False)


def load_json(path):
    with open(path) as f:
        data = json.loads(f.read())
    return data


class AnnotatedObject:
    def __init__(self, d: dict):
        resolved_types = {}
        for cls in self.__class__.mro():
            annotations = getattr(cls, '__annotations__', None)
            if annotations is None:
                continue

            for k, type_ in annotations.items():
                if k in resolved_types:
                    continue

                v = d.get(k, None)
                resolved_types[k] = type_
                if v is None and not self._is_optional_type(type_):
                    # TODO: print warning
                    print(f'{k} is defined in {self.__class__.__name__} but not in {d}')
                    continue
                elif type_ == typing.Union and type(v) not in type_.__args__:
                    # TODO: print warning
                    print(f'{k} is defined in {self.__class__.__name__} but not in {d}')
                    continue
                elif type_ != type(v):
                    # TODO: print warning
                    print(f'{k} is defined in {self.__class__.__name__} as {type_} but {type(v)} in {d}')
                    continue
                if issubclass(type_, AnnotatedObject):
                    v = type_(v)
                setattr(self, k, v)
        self.d = d

    @staticmethod
    def _is_optional_type(type_):
        return typing.Union == type_.__origin__ and len(type_.__args__) == 2 and \
                type(None) == type_.__args__[1]

    def __str__(self):
        return f'[{self.__class__.__name__}] {self.d}'

    @property
    def dict(self):
        return self.d


def get_token_from_config(path):
    try:
        with open(path) as f:
            documents = yaml.full_load(f)
            if not documents:
                raise Exception('Nothing in config file')
            for item, doc in documents.items():
                if item != 'token':
                    raise Exception('No token item')
                elif not doc:
                    raise Exception('Token is not set! Run $savvi init')
                else:
                    return doc
    except EnvironmentError:
        print('Error: Config file not found')


def get_json_headers_with_token(token):
    try:
        authentication = 'token ' + token
    except TypeError:
        print('Token is not set!')
        return

    headers = {
        'Content-Type': 'application/json',
        'Authentication': authentication
    }

    return headers



import requests
import typer
from savvihub.savvihub_cli.constants import DEFAULT_CONFIG_PATH, HOST_URL, CUR_DIR
from savvihub.savvihub_cli.errors import get_error_message
from savvihub.savvihub_cli.utils import *
from savvihub.savvihub_cli.yml_loader import YmlLoader, experiment_conf

experiment_app = typer.Typer()


@experiment_app.callback()
def main():
    """
    Perform your experiment with Savvihub
    """
    return


@experiment_app.command()
def init():
    """
    Initialize a new experiment configuration file
    """
    data = {
        'workspace_slug': typer.prompt('workspace name'),
        'project_slug': typer.prompt('project name'),
        'image_id': int(typer.prompt('image')),
        'resource_spec_id': int(typer.prompt('resource spec')),
        'start_command': typer.prompt('start command'),
    }

    config_file_path = os.path.join(CUR_DIR, "savvihubfile.yml")
    with open(config_file_path, 'w') as f:
        yaml.dump(data, f, default_flow_style=False)

    typer.echo(f"Experiment config successfully made in {config_file_path}")


@experiment_app.command()
def run(
        file: str = typer.Option("savvihubfile.yml", "--file", "-f"),
):
    """
    Run an experiment in Savvihub
    """
    token = get_token_from_config(DEFAULT_CONFIG_PATH)
    headers = get_json_headers_with_token(token)

    yml_loader = YmlLoader(file, experiment_conf)

    # TODO:: dynamically binding the following data fields
    yml_loader.data['branch'] = 'master'
    yml_loader.data['dataset_id_refs'] = {'id': 1, 'ref': 'latest'}

    workspace_slug = yml_loader.pop_val_by_key('workspace_slug')
    project_slug = yml_loader.pop_val_by_key('project_slug')

    res = requests.post(
        url=f'{HOST_URL}/v1/api/workspaces/{workspace_slug}/projects/{project_slug}/experiments/',
        headers=headers,
        data=json.dumps(yml_loader.data),
    )

    res_data = res.json()
    if res.status_code == 400:
        typer.echo(get_error_message(res_data))
        return

    experiment_number = res_data.get('expriment').get('number')
    typer.echo(f"Experiment {experiment_number} is running. Check the experiment status at below link")
    # TODO:: Experiment status URL
    typer.echo(f"URL:")
    return


if __name__ == "__main__":
    experiment_app()



import yaml

experiment_conf = {
    'workspace_slug': str,
    'project_slug': str,
    'image_id': int,
    'resource_spec_id': int,
    'start_command': str,
}


class YmlLoader:
    def __init__(self, filename, conf):
        self.filename = filename
        self.conf = conf
        self.data = self._load()

        if not self.validate():
            raise Exception('Yml load validation error')

    def _load(self):
        try:
            with open(self.filename) as f:
                documents = yaml.full_load(f)
                if not documents:
                    raise Exception('Nothing in config file')
                data_dict = dict()
                for item, doc in documents.items():
                    data_dict[item] = doc
                return data_dict
        except EnvironmentError:
            print('Error: Config file not found')

    def _validate_helper(self, struct, conf):
        if isinstance(struct, dict) and isinstance(conf, dict):
            # struct is a dict of types or other dicts
            return all(k in conf and self._validate_helper(struct[k], conf[k]) for k in struct)
        if isinstance(struct, list) and isinstance(conf, list):
            # struct is list in the form [type or dict]
            return all(self._validate_helper(struct[0], c) for c in conf)
        elif isinstance(conf, type):
            # struct is the type of conf
            return isinstance(struct, conf)
        else:
            # struct is neither a dict, nor list, not type
            return False

    def validate(self):
        return self._validate_helper(self.data, self.conf)

    def pop_val_by_key(self, key):
        return self.data.pop(key, None)






import os
from pathlib import Path


HOST_URL = 'http://localhost:10000'
CUR_DIR = os.getcwd()
DEFAULT_SAVVI_DIR = os.path.join(str(Path.home()), '.savvihub')
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_SAVVI_DIR, 'config.yml')



import os

import grequests
import typer

from savvihub.savvihub_cli.config import config
from savvihub.savvihub_cli.savvihub import SavviHubClient, SavviDatasetFile

dataset_app = typer.Typer()


def parse_dataset_arg(dataset_arg):
    workspace, rest = dataset_arg.split("/")
    if ":" in rest:
        dataset, ref = rest.split(":")
    else:
        dataset = rest
        ref = "latest"
    return workspace, dataset, ref


@dataset_app.callback()
def main():
    return


@dataset_app.command()
def create():
    return


@dataset_app.command()
def push(
    dataset_arg: str = typer.Option(..., "-r"),
    path_arg: str = typer.Argument("."),
):
    workspace, dataset, ref = parse_dataset_arg(dataset_arg)

    client = SavviHubClient(config.token)

    dirs = []
    files = []

    def deepwalk(path):
        for root, dirs_, files_ in os.walk(path_arg):
            for name in dirs_:
                dirs.append(os.path.join(root, name))
                deepwalk(os.path.join(root, name))
            for name in files_:
                files.append(os.path.join(root, name))

    rs = [grequests.post(f'/api/workspaces/{workspace}') for dir in dirs]
    pass


@dataset_app.command()
def pull(
    dataset_arg: str = typer.Option(..., "-r"),
    path_arg: str = typer.Argument("."),
):
    workspace, dataset, ref = parse_dataset_arg(dataset_arg)

    client = SavviHubClient(config.token)
    files = client.dataset_file_list(workspace, dataset, ref=ref)

    def download_file(file: SavviDatasetFile, pg):
        def fn(resp):
            path = os.path.join(path_arg, file.path)
            with open(path, 'wb') as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
            pg.update(1)
        return fn

    with typer.progressbar(length=len(files)) as progress:
        rs = [grequests.get(file.download_url.url, stream=True, callback=download_file(file, progress)) for file in files]
        grequests.map(rs, size=config.parallel)
    typer.echo(f'Downloaded {len(files)} in {os.path.abspath(path_arg)}')


if __name__ == "__main__":
    dataset_app()


import grequests
import requests
import typing

from savvihub.savvihub_cli.constants import HOST_URL
from savvihub.savvihub_cli.utils import AnnotatedObject


class SavviDatasetFile(AnnotatedObject):
    path: str
    is_dir: bool

    size: int
    hash: str

    download_url: str
    upload_url: str


class PaginatedMixin(AnnotatedObject):
    total: int
    startCursor: str
    endCursor: str
    results: typing.List


class SavviHubPaginatedDatasetFilesResponse(PaginatedMixin, AnnotatedObject):
    results: typing.List[SavviDatasetFile]


class SavviHubClient:
    def __init__(self, token, url=HOST_URL, multi=False):
        self.url = url
        self.token = token
        self.headers = {
            'content-type': 'application/json',
        }
        self.api = requests
        if multi:
            self.api = grequests

    def raise_for_status(self, r):
        if r.status_code != 200:
            print(r.text)
            r.raise_for_status()

    def get(self, url, params=None, raw=False):
        r = self.api.get(f'{self.url}{url}', headers=self.headers, params=params)
        self.raise_for_status(r)
        if raw:
            return r.text
        else:
            return r.json()

    def get_all(self, url, params=None):
        raw_resp = self.get(url, params=params)
        resp = PaginatedMixin(raw_resp)
        results = []

        fetched_items = 0
        while True:
            fetched_items += len(resp.results)
            results.extend(resp.results)
            if fetched_items >= resp.total:
                break
            raw_resp = self.get(url, params={**params, 'after': resp.endCursor})
            resp = PaginatedMixin(raw_resp)
        return results

    def post(self, url, data):
        r = self.api.post(f'{self.url}{url}', json=data, headers=self.headers)
        self.raise_for_status(r)
        return r.json()

    def delete(self, url):
        r = self.api.delete(f'{self.url}{url}', headers=self.headers)
        self.raise_for_status(r)
        return r.json()

    def patch(self, url, data):
        r = self.api.patch(f'{self.url}{url}', json=data, headers=self.headers)
        self.raise_for_status(r)
        return r.json()

    def dataset_file_list(self, workspace, dataset, *, ref='latest') -> typing.List[SavviDatasetFile]:
        results = self.get_all(f'/api/workspaces/{workspace}/datasets/{dataset}/files/', params={'ref': ref})
        return [SavviDatasetFile(x) for x in results]

    def dataset_file_create(self, workspace, dataset, path, is_dir):
        r = self.post(f'/api/workspaces/{workspace}/datasets/{dataset}/files/', {
            'path': path,
            'is_dir': is_dir
        })
        return SavviDatasetFile(r)



import requests
import typer

from savvihub.savvihub_cli.config import config
from savvihub.savvihub_cli.constants import HOST_URL, DEFAULT_SAVVI_DIR, DEFAULT_CONFIG_PATH
from savvihub.savvihub_cli.datasets import dataset_app
from savvihub.savvihub_cli.errors import get_error_message
from savvihub.savvihub_cli.experiments import experiment_app
from savvihub.savvihub_cli.utils import *

app = typer.Typer()
app.add_typer(experiment_app, name="experiments")
app.add_typer(dataset_app, name="datasets")
__version__ = '0.0.2'


def version_callback(value: bool):
    if value:
        typer.echo(f"Savvihub CLI Version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
    token: str = typer.Option(None, "--token"),
    parallel: int = typer.Option(20, "--parallel"),
):
    """
    Savvihub command line interface
    """
    if token:
        config.token = token
    if parallel:
        config.parallel = parallel

    return


@app.command()
def ping():
    """
    Ping to server
    """
    res = requests.get(HOST_URL + '/v1/api/ping/')
    typer.echo(f"Response code: {res.status_code}, Response text: {res.text}")


@app.command()
def login():
    """
    Login command
    """
    email = typer.prompt("Email")
    password = typer.prompt("Password")
    data = {
        'email': email,
        'passowrd': password,
    }

    headers = {'Content-Type': 'application/json'}
    res = requests.post(
        url=f'{HOST_URL}/v1/api/accounts/signin/',
        headers=headers,
        data=json.dumps(data),
    )

    res_data = res.json()
    if res.status_code == 400:
        typer.echo(get_error_message(res_data))
        return

    token = res_data.get('token')
    typer.echo("Initialize with this token")
    typer.echo(f"token: {token}")
    return


@app.command()
def init():
    """
    Initialize with token
    """
    data = {
        'token': typer.prompt('token')
    }

    make_dir(DEFAULT_SAVVI_DIR)
    with open(DEFAULT_CONFIG_PATH, 'w') as f:
        yaml.dump(data, f, default_flow_style=False)

    typer.echo(f"Token successfully saved in {DEFAULT_CONFIG_PATH}")
